// Laundrymon
// Wireless, non-invasive laundry monitor
// 2009 Lucas Eckels
// http://lucaseckels.com/laundrymon

#include <Wire.h>
#include <WProgram.h>
#include <EEPROM.h>
#include "nunchuck_funcs.h"
#include "AtCommands.h"

//#define ALWAYS_DEBUG

#define MAGIC "LM0"
#define MAGIC_LENGTH 3
#define ROLLING_LENGTH 5
#define MAX_BYTE 255
#define DEVICE_COUNT 2

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))

struct PersistentSettings {
  char magic[MAGIC_LENGTH];
  
  byte loopDelay; // in 100ms increments
  
  unsigned int threshold[DEVICE_COUNT];
  byte onCount[DEVICE_COUNT]; // Number of times it must be above the threshold to be 'on'
  byte offCount[DEVICE_COUNT]; // Number of times it must be below the threshold to be 'off'
} persistentSettings;
  
#define LOOP_DELAY_DEFAULT 10;

#define WASHER_THRESHOLD_DEFAULT (25 * ROLLING_LENGTH)
#define WASHER_ON_COUNT_DEFAULT 10
#define WASHER_OFF_COUNT_DEFAULT 100

#define DRYER_THRESHOLD_DEFAULT (25 * ROLLING_LENGTH)
#define DRYER_ON_COUNT_DEFAULT 10
#define DRYER_OFF_COUNT_DEFAULT 100

#define WASHER_PIN 14
#define DRYER_PIN 15

#define DEBUG_OUTPUT_ON_PERIOD 100

#ifdef ALWAYS_DEBUG
byte debugLevel = MAX_BYTE;
#else
bool debugLevel = 0;
#endif
byte debugOutputOn = 0;  // How many iterations debug has been on

#define W_ 0
#define D_ 1

#define ON true
#define OFF false

char *names[DEVICE_COUNT] = {"Washer", "Dryer"};

struct State {
  int lastX;
  int lastY;
  int lastZ;
  
  int totalVibe;
  byte vibes[ROLLING_LENGTH];
  byte vibePos;
  byte onCount;
  byte offCount;
  
  bool state;
  
  State() :
      lastX(0),
      lastY(0),
      lastZ(0),
      totalVibe(0),
      vibePos(0),
      onCount(0),
      offCount(0),
      state(OFF) {
    memset(vibes, 0, sizeof(vibes));
  }
     
  // Returns the total vibration    
  int AddValue(int x, int y, int z) {   
    int dx = abs(x - lastX);
    int dy = abs(y - lastY);
    int dz = abs(z - lastZ);
    totalVibe -= vibes[vibePos];
    int newVibe = dx + dy + dz;
 
    if (debugLevel) {
      Serial.print("\t");
      Serial.print(x, DEC);
      Serial.print(", ");
      Serial.print(y, DEC);
      Serial.print(", ");
      Serial.print(z, DEC);
      Serial.print(" - ");
      Serial.print(newVibe, DEC);
      Serial.print("\r\n");
    }
 
    vibes[vibePos] = MIN(newVibe, MAX_BYTE);
    totalVibe += vibes[vibePos];
    vibePos = (vibePos + 1) % ROLLING_LENGTH;
    lastX = x;
    lastY = y;
    lastZ = z;
 
    if (debugLevel > 5) {
      Serial.print("Vibes: ");
      for (byte i = 0; i < 5; ++i) {
        Serial.print(vibes[i], DEC);
        Serial.print(", ");
      }
      Serial.print(" Total: ");
      Serial.print(totalVibe, DEC);
      Serial.print("\r\n");
    } 
    return totalVibe;
  }
  
  byte IncrementOn() {
    offCount = 0;
    if (state == OFF) {
      ++onCount;
    }
    return onCount;
  }
  
  byte IncrementOff() {
    onCount = 0;
    if (state == ON) {
      ++offCount;
    }
    return offCount;
  }
  
  bool setState(bool newState) {
    if (newState != state) {
      state = newState;
      return true;
    }
    return false;
  }
  
};

State states[DEVICE_COUNT];

void setActive(byte device) {
  if (device == D_) {
    digitalWrite(WASHER_PIN, LOW);
    digitalWrite(DRYER_PIN, HIGH);
  } else if (device == W_) {
    digitalWrite(DRYER_PIN, LOW);
    digitalWrite(WASHER_PIN, HIGH);
  }
  delay(10);
}

void readEeprom(void *buffer, size_t length) {
  byte *b = (byte*)buffer;
  for (size_t i = 0; i < length; ++i) {
    b[i] = EEPROM.read(i);
  }
}

void writeEeprom(void *buffer, size_t length) {
  byte *b = (byte*)buffer;
  for (size_t i = 0; i < length; ++i) {
    EEPROM.write(i, b[i]);
  }
}

void reset() {
  memcpy(persistentSettings.magic, MAGIC, MAGIC_LENGTH);
  persistentSettings.loopDelay = LOOP_DELAY_DEFAULT;
  persistentSettings.threshold[W_] = WASHER_THRESHOLD_DEFAULT;
  persistentSettings.onCount[W_] = WASHER_ON_COUNT_DEFAULT;
  persistentSettings.offCount[W_] = WASHER_OFF_COUNT_DEFAULT;
  persistentSettings.threshold[D_] = DRYER_THRESHOLD_DEFAULT;
  persistentSettings.onCount[D_] = DRYER_ON_COUNT_DEFAULT;
  persistentSettings.offCount[D_] = DRYER_OFF_COUNT_DEFAULT;
}

void setup() {
  pinMode(WASHER_PIN, OUTPUT);
  pinMode(DRYER_PIN, OUTPUT);
  digitalWrite(WASHER_PIN, HIGH);
  digitalWrite(DRYER_PIN, HIGH);
  
  Serial.begin(9600);
  delay(100);
  nunchuck_init();
  delay(1000);

  readEeprom(&persistentSettings, sizeof(PersistentSettings));
  if (memcmp(persistentSettings.magic, MAGIC, MAGIC_LENGTH)) {
    reset();
  }

  Serial.print("Laundrymon ready\r\n");  
}

void CommandReset(const char *command, char *buffer, byte length) {
  reset();
  atCommandOk();
}

void CommandWrite(const char *command, char *buffer, byte length) {
  writeEeprom(&persistentSettings, sizeof(PersistentSettings));
  atCommandOk();
}

void CommandDebug(const char *command, char *buffer, byte length) {
  if (length < 1) {
    atCommandError("Expected argument");
    return;
  }
  debugLevel = atoi(buffer);
  Serial.print("Debug level set to ");
  Serial.print(debugLevel, DEC);
  Serial.print("\r\n");
  atCommandOk();
}

void CommandPrintSettings(const char *command, char *buffer, byte length) {
  Serial.print("Loop delay: ");
  Serial.print(persistentSettings.loopDelay, DEC);
  Serial.print(" * 100ms\r\n");
  for (byte i = 0; i < DEVICE_COUNT; ++i) {
    Serial.print(names[i]);
    Serial.print("\r\n  Threshold: ");
    Serial.print(persistentSettings.threshold[i], DEC);
    Serial.print("\r\n  On count: ");
    Serial.print(persistentSettings.onCount[i], DEC);
    Serial.print("\r\n  Off count: ");
    Serial.print(persistentSettings.offCount[i], DEC);
    Serial.print("\r\n");
  }
  atCommandOk();
}

void CommandSetDeviceSetting(const char *command, char *buffer, byte length) {
  if (debugLevel) {
    Serial.print("Buffer: ");
    Serial.print(buffer);
    Serial.print("\r\n");
  }
  byte device = -1;
  if (command[0] == 'W') {
    device = W_;
  } else if (command[0] == 'D') {
    device = D_;
  } else {
    atCommandError("unexpected device");
    return;
  }
  int val = atoi(buffer);
  if (debugLevel) {
    Serial.print("Value: ");
    Serial.print(val, DEC);
    Serial.print("\r\n");
  }
  
  if (command[1] == 'N') {
    persistentSettings.onCount[device] = val;
  } else if (command[1] == 'F') {
    persistentSettings.offCount[device] = val;
  } else if (command[1] == 'T') {
    persistentSettings.threshold[device] = val;
  } else {
    atCommandError("unknown command");
  }
  atCommandOk();
}

void CommandSetLoopDelay(const char *command, char *buffer, byte length) {
  int val = atoi(buffer);
  if (val == 0) {
    atCommandError("Bad argument");
  }
  persistentSettings.loopDelay = val;
  atCommandOk();
}


AT_COMMAND_MAP_BEGIN(Laundrymon)
AT_COMMAND_MAP_ENTRY("RE", CommandReset);
AT_COMMAND_MAP_ENTRY("WR", CommandWrite);
AT_COMMAND_MAP_ENTRY("DE", CommandDebug);
AT_COMMAND_MAP_ENTRY("PR", CommandPrintSettings);
AT_COMMAND_MAP_ENTRY("WN", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("WF", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("WT", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("DN", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("DF", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("DT", CommandSetDeviceSetting);
AT_COMMAND_MAP_ENTRY("LD", CommandSetLoopDelay);
AT_COMMAND_MAP_END(Laundrymon)

#define DEVICES_CONNECTED

void loop() {
  atCommandHandlerLaundrymon();
  
#ifdef DEVICES_CONNECTED
  for (byte i = 0; i < DEVICE_COUNT; ++i) {
    if (debugLevel) {
      Serial.print(names[i]);
    }
    setActive(i);
    nunchuck_get_data();
    int val = states[i].AddValue(nunchuck_accelx(),
                                 nunchuck_accely(),
                                 nunchuck_accelz());
    if (val > persistentSettings.threshold[i]) {
      if (states[i].IncrementOn() > persistentSettings.onCount[i]) {
        if (states[i].setState(ON)) {
          Serial.print(names[i]);
          Serial.print(" turned on\r\n");
        }
      }
    } else {
      if (states[i].IncrementOff() > persistentSettings.offCount[i]) {
        if (states[i].setState(OFF)) {
          Serial.print(names[i]);
          Serial.print(" turned off\r\n");
        }
      }
    }
  }
#endif
  
  #ifndef ALWAYS_DEBUG
  if (debugLevel && ++debugOutputOn >= DEBUG_OUTPUT_ON_PERIOD) {
    Serial.print("Debug timing out\r\n");
    debugLevel = false;
  }  
  #endif
  
  delay(persistentSettings.loopDelay * 100);
}

